/**
Copyright © 2023 Sebastian Roith <https://www.roith-unterrichtet.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

/***[ JavaCodeReader.js ] *************************/
/** Class converting unicode sequences and managing code position */
function JavaCodeReader(
    code        // Javalike code to read
) {
    this.code = code;
    if (code.length && code.charCodeAt(code.length - 1) == 26)
        this.code = code.substring(0, code.length - 1);
}

/** Javalike code to read */
JavaCodeReader.prototype.code = '';

/** Present position within this.code */
JavaCodeReader.prototype.offset = 0;

/** Present line number (starting at 1) */
JavaCodeReader.prototype.line = 1;

/** Present position in present line (starting at 1) */
JavaCodeReader.prototype.pos = 1;

/** Line number of present token's start (starting at 1) */
JavaCodeReader.prototype._tokenLine = 1;

/** Position in line of present token's start (starting at 1) */
JavaCodeReader.prototype._tokenPos = 1;

/** Position within this.code of present token's start */
JavaCodeReader.prototype._tokenOffset = 0;

/**
* Get next char
* @return String        Next char (or "")
*/
JavaCodeReader.prototype.getNextChar = function() {
    return this.code.substring(this.offset, this.offset + 1);
}

/**
* Get char at given offset
* @return String        Requested char (or "")
*/
JavaCodeReader.prototype.getByOffset = function(
    offset          // Offset
) {
    return this.code.substring(this.offset + offset, this.offset + 1 + offset);
}

/**
* Get next substring of code beeing `length` chars long
* @return String        Substring
*/
JavaCodeReader.prototype.getNextString = function(
    length          // Requested length
) {
    return this.code.substring(this.offset, this.offset + length);
}

/**
* Check if next substring of code starts with `target`
* @return Boolean       Does remaining part of code start with `target`?
*/
JavaCodeReader.prototype.compareNext = function(
    target          // String to compare
) {
    return (this.getNextString(target.length) == target);
}

/**
* Step `length` chars forward in this.code
* @return JavaCodeReader        Value of `this`
*/
JavaCodeReader.prototype.forward = function(
    length          // Optional: Number of chars to step forward (defaults to 1)
) {
    if (!Number.isInteger(length) || length < 1)
        length = 1;
    for (let n = 0; n < length; ++n) {
        if (this.getNextChar() == '\n') {
            ++this.line;
            this.pos = 1;
        } else
            ++this.pos;
        ++this.offset;
    }
    return this;
}

/**
* Get next char of code and forward code by one char
* @return String        Next char (or "")
*/
JavaCodeReader.prototype.consumeNextChar = function() {
    const nextChar = this.getNextChar();
    this.forward();
    return nextChar;
}

/**
* Add information on position to present token
* @return undefined
*/
JavaCodeReader.prototype.addPositionInfo = function(
    token       // JavaToken instance to add information to
) {
    token.line = this._tokenLine;
    token.pos = this._tokenPos;
    const realCode = this.code.substring(this._tokenOffset, this.offset);
    if (token.value !== realCode)
        token.raw = realCode;
    this._tokenLine = this.line;
    this._tokenPos = this.pos;
    this._tokenOffset = this.offset;
}

/***[ JavaToken.js ] *************************/
/** A single token in Javalike code */
function JavaToken(
    type,           // Type of token (see JavaToken.TYPE_*)
    value,          // Value of token
    reader          // Responsible JavaCodeReader instance
) {
    this.type = type;
    this.value = value;
    reader.addPositionInfo(this);
}

/** Type of token (see JavaToken.TYPE_*) */
JavaToken.prototype.type = -1;

/** Value of token */
JavaToken.prototype.value = undefined;

/** Actual code producing token */
JavaToken.prototype.raw = undefined;

/** Line in with code starts (starting at 1) */
JavaToken.prototype.line = 0;

/** Number of char within line (starting at 1) of token's start */
JavaToken.prototype.pos = 0;

/**
* Convert token to java code
* @return String    Java code represented by this token
*/
JavaToken.prototype.toString = function() {
    function escape(s, quote) {
        if (s == '"' && quote == "'")
            return "'\"'";
        const escaped = JSON.stringify(s);
        if (quote == '"')
            return escaped;
        return quote + escaped.substring(1, escaped.length - 1) + quote;
    }

    if ([JavaToken.TYPE_LITERAL_STRING, JavaToken.TYPE_LITERAL_CHAR].indexOf(this.type) != -1)
        return escape(this.value, (this.type == JavaToken.TYPE_LITERAL_STRING) ? '"' : "'");
    return '' + this.value;
}

/** Token types */
JavaToken.TYPE_ERROR = 0;
JavaToken.TYPE_WHITESPACE = 1;
JavaToken.TYPE_COMMENT = 2;
JavaToken.TYPE_IDENTIFIER = 3;
JavaToken.TYPE_KEYWORD = 4;
JavaToken.TYPE_SEPARATOR = 5;
JavaToken.TYPE_OPERATOR = 6;
JavaToken.TYPE_LITERAL_NULL    = 10;
JavaToken.TYPE_LITERAL_INT     = 11;
JavaToken.TYPE_LITERAL_LONG    = 12;
JavaToken.TYPE_LITERAL_FLOAT   = 13;
JavaToken.TYPE_LITERAL_DOUBLE  = 14;
JavaToken.TYPE_LITERAL_BOOLEAN = 15;
JavaToken.TYPE_LITERAL_CHAR    = 16;
JavaToken.TYPE_LITERAL_STRING  = 17;

/***[ JavaLexer.js ] *************************/
/** Class to generate tokens of Javalike code */
function JavaLexer(
    code        // Javalike code to lex
) {
    const reader = new JavaCodeReader(code);
    this.tokens = [];
    let token;
    while (token = this._tokenCreators.reduce((prev, fktCreate) => prev || fktCreate.call(reader), null))
        this.tokens.push(token);
}

/** Tokens found by lexer */
JavaLexer.prototype.tokens = [];

/**
* Read a char or string literal
* @return JavaToken     Created token
*/
JavaLexer._readString = function(
    reader,     // JavaCodeReader instance to use
    delimiter   // Delimiter char of string
) {
    let length = 1;
    let data = '';
    let escaped = false;
    let error = null;
    while (true) {
        let next = reader.getByOffset(length);
        ++length;
        if (!error && !next.length)
            error = 'Zugehöriges Endzeichen ' + delimiter + ' fehlt';
        if (!error && next == '\n') {
            error = 'Zugehöriges Endzeichen ' + delimiter + ' muss vor dem Zeilenende stehen';
            next = '';
            --length;
        }
        if (!next.length || (!escaped && next == delimiter)) {
            reader.forward(length);
            if (!error && delimiter == "'" && (data.length != 1))
                error = 'Zwischen einfachen Anführungszeichen muss genau ein Zeichen stehen';
            if (error)
                return new JavaToken(JavaToken.TYPE_ERROR, error, reader);
            return new JavaToken((delimiter == "'") ? JavaToken.TYPE_LITERAL_CHAR : JavaToken.TYPE_LITERAL_STRING, data, reader);
        }
        if (escaped) {
            escaped = false;
            switch (next) {
                case "'":
                case '"':
                case '\\':
                    data += next;
                    break;
                case 'b':
                    data += String.fromCharCode(8);
                    break;
                case 't':
                    data += String.fromCharCode(9);
                    break;
                case 'n':
                    data += String.fromCharCode(10);
                    break;
                case 'f':
                    data += String.fromCharCode(12);
                    break;
                case 'r':
                    data += String.fromCharCode(13);
                    break;
                case 'u':
                    let hex = '';
                    while (hex.length < 4) {
                        const hexChar = reader.getByOffset(length);
                        if (!hexChar.length || ('0123456789abcdefABCDEF'.indexOf(hexChar) == -1)) {
                            if (!error)
                                error = 'Nach \\u muss eine vierstellige hexadezimale Zahl stehen';
                            break;
                        }
                        ++length;
                        hex += hexChar;
                    }
                    data += String.fromCharCode(parseInt(hex, 16));
                    break;
                default:
                    if (!error)
                        error = 'Die Escape-Sequenz \\' + next + ' gibt es nicht';
                    break;
            }
            continue;
        }
        if (next == '\\') {
            escaped = true;
            continue;
        }
        data += next;
    }
}

/**
* If possible: Read decimal number literal and create token
* @return JavaToken     Created token (or null)
*/
JavaLexer._readDecimalNumber = function(
    reader      // JavaCodeReader instance to use
) {
    let number = reader.getNextChar();
    let length = 1;
    let underscores = 0;
    if (!number.length || '0123456789.'.indexOf(number) == -1)
        return null;
    while (true) {
        const next = reader.getByOffset(length);
        if (next == '.' && number.indexOf('.') != -1)
            break;
        if (next == '.' && underscores)
            break;
        if (next == '_' && number.charAt(number.length -1) == '.')
            break;
        if (!underscores && (next == 'e' || next == 'E')) {
            const lengthBeforeE = length;
            ++length;
            let negative = false;
            let exponent = reader.getByOffset(length);
            if (exponent == '-') {
                negative = true;
                ++length;
            }
            if (exponent == '+')
                ++length;
            exponent = '';
            while (true) {
                const exponentChar = reader.getByOffset(length);
                if (exponentChar == '_' && !exponent.length)
                    break;
                if (!exponentChar.length || '0123456789'.indexOf(exponentChar) == -1)
                    break;
                ++length;
                if (exponentChar == '_') {
                    ++underscores;
                    continue;
                }
                exponent += exponentChar;
                underscores = 0;
            }
            if (!exponent.length) {
                length = lengthBeforeE;
                break;
            }
            length -= underscores;

            const typeIndicator = underscores ? '' : reader.getByOffset(length).toUpperCase();
            let type = JavaToken.TYPE_LITERAL_FLOAT;
            if (typeIndicator == 'F')
                ++length;
            if (typeIndicator == 'D') {
                ++length;
                type = JavaToken.TYPE_LITERAL_DOUBLE;
            }
            return new JavaToken(type, parseFloat(number, 10) * Math.pow(10, (negative ? -1 : 1) * parseInt(exponent, 10)), reader.forward(length));
        }
        if (!next.length || '0123456789._'.indexOf(next) == -1)
            break;
        ++length;
        if (next == '_') {
            ++underscores;
            continue;
        }
        number += next;
        underscores = 0;
    }
    length -= underscores;
    if (number == '.')
        return null;

    if (number.indexOf('.') == -1) {
        const typeIndicator = underscores ? '' : reader.getByOffset(length).toUpperCase();
        let type = JavaToken.TYPE_LITERAL_INT;
        if (typeIndicator == 'L') {
            ++length;
            type = JavaToken.TYPE_LITERAL_LONG;
        }
        if (typeIndicator == 'F') {
            ++length;
            type = JavaToken.TYPE_LITERAL_FLOAT;
        }
        if (typeIndicator == 'D') {
            ++length;
            type = JavaToken.TYPE_LITERAL_DOUBLE;
        }
        return new JavaToken(type, parseInt(number, 10), reader.forward(length));
    }

    const typeIndicator = underscores ? '' : reader.getByOffset(length).toUpperCase();
    let type = JavaToken.TYPE_LITERAL_FLOAT;
    if (typeIndicator == 'F')
        ++length;
    if (typeIndicator == 'D') {
        ++length;
        type = JavaToken.TYPE_LITERAL_DOUBLE;
    }
    return new JavaToken(type, parseFloat(number, 10), reader.forward(length));
}

/**
* Create a function creating tokens to fixed key words
* @return Function      Function eventually creating next token
*/
JavaLexer._fixedWordToToken = function(
    arrWord,
    type,
    noAutoTermination
) {
    return function() {
        return arrWord.reduce((prev, word) => {
            if (prev)
                return prev;
            if (!this.compareNext(word))
                return null;
            if (noAutoTermination) {
                const nextChar = this.getByOffset(word.length);
                if (nextChar.trim().length && '><=!&|+-*/^%!?:.,;()[]{}'.indexOf(nextChar) == -1)
                    return null;
            }
            return new JavaToken(type, word, this.forward(word.length));
        }, null);
    };
};

/** Functions to possibly create next token */
JavaLexer.prototype._tokenCreators = [
    // Whitespace?
    function() {
        const start = this.getNextChar();
        if (!start.length || start.trim().length)
            return null;
        let length = 1;
        for (let next = this.getByOffset(1); next.length && !next.trim().length; next = this.getByOffset(length))
            ++length;
        return new JavaToken(JavaToken.TYPE_WHITESPACE, this.getNextString(length), this.forward(length));
    },
    // Line comment?
    function() {
        if (!this.compareNext('//'))
            return null;
        let length = 2;
        for (let next = this.getByOffset(length); next.length && next != '\n'; next = this.getByOffset(length))
            ++length;
        return new JavaToken(JavaToken.TYPE_COMMENT, this.getNextString(length), this.forward(length));
    },
    // Block comment?
    function() {
        if (!this.compareNext('/*'))
            return null;
        let length = 2;
        let foundStar = true;
        for (let next = this.getByOffset(length); next.length; next = this.getByOffset(length)) {
            ++length;
            if (next == '*') {
                foundStar = true;
                continue;
            }
            if (foundStar && next == '/')
                break;
            foundStar = false;
        }
        return new JavaToken(JavaToken.TYPE_COMMENT, this.getNextString(length), this.forward(length));
    },
    // Numberic literal?
    function() {
        return JavaLexer._readDecimalNumber(this);
    },
    // Fixed substrings?
    JavaLexer._fixedWordToToken(['>>>=', '<<=', '>>=', '>>>', '==', '<=', '>=', '!=', '&&', '||', '++', '--', '<<', '>>', '+=', '-=', '*=', '/=', '&=', '|=', '^=', '%=', '=', '>', '<', '!', '~', '?', ':', '+', '-', '*', '/', '&', '|', '^', '%'], JavaToken.TYPE_OPERATOR),
    JavaLexer._fixedWordToToken(['(', ')', '{', '}', '[', ']', ';', ',', '.'], JavaToken.TYPE_SEPARATOR),
    JavaLexer._fixedWordToToken(['null'], JavaToken.TYPE_LITERAL_NULL, true),
    JavaLexer._fixedWordToToken(['abstract', 'continue', 'for', 'new', 'switch', 'assert', 'default', 'if', 'package', 'synchronized', 'boolean', 'do', 'goto', 'private', 'this', 'break', 'double', 'implements', 'protected', 'throw', 'byte', 'else', 'import', 'public', 'throws', 'case', 'enum', 'instanceof', 'return', 'transient', 'catch', 'extends', 'int', 'short', 'try', 'char', 'final', 'interface', 'static', 'void', 'class', 'finally', 'long', 'strictfp', 'volatile', 'const', 'float', 'native', 'super', 'while'], JavaToken.TYPE_KEYWORD, true),
    // Boolean literal?
    function() {
        if (this.compareNext('true'))
            return new JavaToken(JavaToken.TYPE_LITERAL_BOOLEAN, true, this.forward(4));
        if (this.compareNext('false'))
            return new JavaToken(JavaToken.TYPE_LITERAL_BOOLEAN, false, this.forward(5));
        return null;
    },
    // null?
    function() {
        if (this.compareNext('null'))
            return new JavaToken(JavaToken.TYPE_LITERAL_NULL, null, this.forward(4));
        return null;
    },
    // String or char?
    function() {
        const delimiter = this.getNextChar();
        if (delimiter != '"' && delimiter != "'")
            return null;
        return JavaLexer._readString(this, delimiter);
    },
    // Create identifier?
    function() {
        let identifier = this.consumeNextChar();
        if (!identifier.length)
            return null;
        const endChars = '=<>!~?:&|+-*/%^(){}[];,."\'';
        for (let next = this.getNextChar(); endChars.indexOf(next) == -1 && next.trim().length; next = this.getNextChar())
            identifier += this.consumeNextChar();
        return new JavaToken(JavaToken.TYPE_IDENTIFIER, identifier, this);
    }
];

/***[ JavaPart.js ] *************************/
/** Abstract parent class of all possible parts within a Java-like source code */
function JavaPart(
    tokenStart      // First token belonging to part
) {
    this.tokenStart = tokenStart;
    this.tokenEnd = tokenStart;
}

/** Start token of part */
JavaPart.prototype.tokenStart = null;

/** Last token of part */
JavaPart.prototype.tokenEnd = null;

/**
* Check if single attribute value fits target value
* @return boolean       Does it fit?
*/
JavaPart.prototype.compareSingleValue = function(
    key,            // Name of attribute to check
    target          // Value (or array of possible values) to compare
) {
    if (target === undefined)
        return true;
    if (this[key] instanceof Array) {
        function cmpArrays(arr1, arr2) {
            if (arr1.length != arr2.length)
                return false;
            arr1 = arr1.slice();
            arr2 = arr2.slice();
            arr1.sort((a, b) => (a < b) ? -1 : ((a > b) ? 1 : 0));
            arr2.sort((a, b) => (a < b) ? -1 : ((a > b) ? 1 : 0));
            return arr1.reduce((prev, value, index) => (prev && (value == arr2[index])), true);
        }
        if (!(target instanceof Array))
            return false;
        if (target.length && target.reduce((prev, item) => (prev && (item instanceof Array)), true))
            return target.reduce((prev, target) => (prev || cmpArrays(this[key], target)), false);
        return cmpArrays(this[key], target);
    }
    if (target instanceof Array)
        return (target.indexOf(this[key]) != -1);
    return (target == this[key]);
}

/**
* Find indizes of minimal substring of `actual` that has to be changed to make `actual` match `target`
* @return Array     [ 'Index of first not matching element of `actual`', ' Index of first matching element of `actual` follwing not matching elements' ] or `false` if arrays match completely
*/
JavaPart.findMissmatchSubarray = function(
    actual,     // Actual array
    target,     // Array describing wanted state
    fktIsEqual  // Function like (a, b) => a == b to check equality of two items
) {
    const firstMismatchIndex = actual.reduce((prev, curr, index) => {
        if (prev !== false)
            return prev;
        if (index >= target.length || !fktIsEqual(target[index], curr))
            return index;
        return false;
    }, false);
    if (firstMismatchIndex === false)
        return (actual.length == target.length) ? false : [actual.length, actual.length];
    let lastMatchIndex = actual.length - 1;
    while (lastMatchIndex >= firstMismatchIndex && fktIsEqual(actual[lastMatchIndex], target[target.length + lastMatchIndex - actual.length]))
        --lastMatchIndex;
    ++lastMatchIndex;
    return [firstMismatchIndex, lastMatchIndex];
}

/**
* If relevant, customize the source code to this object
* @return String    Customized code
*/
JavaPart.prototype.adaptTargetCode = function(
    code
) {
    return code;
}

/**
* Compares plain content (just consisting of tokens without whitespace and so on) with code
* @return Object    { conformity: "Proportion of values equaling target values", problems: "Array of mismatches", instance: "First mismatching token" }
*/
JavaPart.prototype.comparePlainContent = function(
    target      // String representing wanted code or array containing alternative code strings (comments and whitespace is stripped); or `true` (any content is okay) or `false` (there must be no content)
) {
    if (target instanceof Array) {
        if (!target.length)
            return this.comparePlainContent(false);
        return target.map(t => this.comparePlainContent(t))
            .reduce((prev, next) => (!prev || next.conformity > prev.conformity) ? next : prev, null);
    }

    const createReportablePart = arrContent => {
        if (!(arrContent instanceof Array) || !arrContent.length)
            return this;
        const part = new JavaPart(arrContent[0]);
        part.tokenEnd = arrContent[arrContent.length - 1];
        return part;
    };

    // All cases in which it is only specified whether there must be content
    const contentSpecified = (typeof target == 'string' || target instanceof String);
    if (!contentSpecified && (this.content !== null) == target)
        return { conformity: 1, problems: [], instance: null };
    if (!contentSpecified && !target)
        return {
            conformity: 0,
            problems: [(this instanceof JavaPartAttributeDeclaration) ? 'Dieses Attribut muss ohne Wertezuweisung deklariert werden' : 'Es darf hier keine innere Sequenz geben'],
            instance: createReportablePart(this.content)
        };
    if (target && (!contentSpecified || this.content == null))
        return { conformity: 0, problems: [(this instanceof JavaPartAttributeDeclaration) ? 'Wertezuweisung fehlt' : 'Innere Sequenz fehlt'], instance: this };

    // Now we can assume, target specifies a single code
    const tokens = (new JavaLexer(this.adaptTargetCode(target))).tokens;
    const [errorToken] = tokens.filter(token => !token.type);
    if (errorToken) {
        if (typeof console == 'object' && typeof console.warn == 'function')
            console.warn("Syntaxfehler in Vergleichscode `" + target + "`: „" + errorToken.value + "“ nicht erkannt", errorToken);
        return { conformity: 0, instance: this, problems: ["Syntaxfehler in Vergleichscode `" + target + "`: „" + errorToken.value + "“ nicht erkannt"]};
    }
    const relevantTokens = tokens.filter(token => ([JavaToken.TYPE_COMMENT, JavaToken.TYPE_WHITESPACE, JavaToken.TYPE_ERROR].indexOf(token.type) == -1));

    const missmatch = JavaPart.findMissmatchSubarray(
        this.content,
        relevantTokens,
        (a, b) => (typeof a == 'object' && typeof b == 'object') ? (a.value == b.value) : false
    );
    if (!missmatch)
        return { conformity: 1, problems: [], instance: null };
    const result = {};
    result.conformity = 1 - Math.max(((missmatch[1] - missmatch[0]) / Math.max(1, this.content.length) + (relevantTokens.length + missmatch[1] - this.content.length - missmatch[0]) / Math.max(1, relevantTokens.length)) / 2, 0.0001);
    if (missmatch[0] >= missmatch[1]) {
        result.problems = ['An der markierten Stelle fehlt etwas'];
        result.instance = createReportablePart(this.content.slice(
            Math.max(0, Math.min(missmatch[0], this.content.length) - 1),
            Math.max(1, Math.min(missmatch[1] + 1, this.content.length))
        ));
    } else {
        result.problems = ['Das Markierte ist fehlerhaft'];
        result.instance = createReportablePart(this.content.slice(missmatch[0], missmatch[1]));
    }
    return result;
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPart.prototype.toJSON = function() {
    return null;
}

/**
* Compare an array of target values' objects with array of actual JavaPart instances
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPart.compareContent = function(
    actual,
    target
) {
    if (target.length && !target.reduce((prev, item) => prev + (item.part ? 1 : 0), 0))
        return target
            .map(item => JavaPart.compareContent(actual, item))
            .reduce((a, b) => (b.conformity > a.conformity) ? b : a);

    actual = actual.slice();
    target = target.slice();

    /**
    * Compares `goodFit` value and eventually conformity of a and b
    * @return Number        Negative, if `a` is better, positive, if `b` is better and 0 otherwise
    */
    function cmpConformity(a, b) {
        if (a.goodFit && !b.goodFit)
            return -1;
        if (!a.goodFit && b.goodFit)
            return +1;
        return b.conformity - a.conformity;
    }

    function calcBestConformity(actualItem) {
        return target
            .map(targetItem => {
                const result = actualItem.compareWithTargetValues(targetItem);
                result._actual = actualItem;
                result._target = targetItem;
                return result;
            }).reduce((a, b) => (cmpConformity(a, b) <= 0) ? a : b, { _actual: actualItem, _target: null, conformity: 0, goodFit: false, instance: actualItem, problems: [actualItem + ' darf es nicht geben'] });
    }

    let checked = 0;
    let matches = 0;
    let instance = null;
    const problems = [];
    let conformities = actual.map(actualItem => calcBestConformity(actualItem));
    while (actual.length && target.length) {
        conformities.sort((a, b) => cmpConformity(a, b));
        const bestMatch = conformities.shift();
        const actualItem = bestMatch._actual;
        const targetItem = bestMatch._target;

        ++checked;
        if (bestMatch.problems.length) {
            bestMatch.problems.forEach(problem => problems.push(problem));
            instance = instance || bestMatch.instance;
            matches += bestMatch.conformity;
        } else
            ++matches;

        actual = actual.filter(item => (item != actualItem));
        target = target.filter(item => (item != targetItem));
        conformities.forEach((obj, index, arr) => {
            if (obj._target == targetItem)
                arr[index] = calcBestConformity(obj._actual);
        });
    }
    actual.forEach(actualItem => {
        problems.push(actualItem + ' darf es nicht geben');
        instance = instance || actualItem;
        ++checked;
    });
    target.map(targetItem => {
        ++checked;
        switch (targetItem.part) {
            case 'class':
                if (targetItem.name)
                    return 'Klasse ' + targetItem.name + ' fehlt';
                return 'Klassendefinition fehlt';
            case 'attribute':
                if (targetItem.name)
                    return 'Attribut ' + targetItem.name + ' fehlt';
                return 'Attribut fehlt';
            case 'constructor':
                return 'Konstruktor fehlt';
            case 'method':
                if (targetItem.name)
                    return 'Methode ' + targetItem.name + ' fehlt';
                return 'Methode fehlt';
            /*!*/
            default:
                return 'Etwas fehlt noch';
        }
    }).forEach(problem => problems.push(problem));

    return { conformity: matches / checked, instance, problems, goodFit: false };
}

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPart.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    return { conformity: 0, goodFit: false, problems: ['Abgleich noch nicht von Lernpaket implementiert'], instance: this };
}

/** Types of visibility */
JavaPart.VISIBILITY_PRIVATE = 0;
JavaPart.VISIBILITY_PROTECTED = 1;
JavaPart.VISIBILITY_PACKAGE = 2;
JavaPart.VISIBILITY_PUBLIC = 3;

/**
* Read visibility declaration from tokens if available
* @return Array     ["JavaPart.VISIBILITY_* value or undefined (if there is no visibility declaration)", "Index of next token to read"]
*/
JavaPart.visibilityFromTokens = function(
    tokens,     // Array of tokens (without comments and whitespace)
    offset      // Optional: Index tokens of token in question
) {
    const index = offset ? offset : 0;
    if (index >= tokens.length || tokens[index].type != JavaToken.TYPE_KEYWORD)
        return [undefined, index];
    switch (tokens[index].value) {
        case 'private':
            return [JavaPart.VISIBILITY_PRIVATE, index + 1];
        case 'protected':
            return [JavaPart.VISIBILITY_PROTECTED, index + 1];
        case 'public':
            return [JavaPart.VISIBILITY_PUBLIC, index + 1];
    }
    return [undefined, index];
}

/**
* Read variable type from tokens if available
* @return Array     ["Variable type or undefined (if there is no type)", "Index of next token to read"]
*/
JavaPart.typeFromTokens = function(
    tokens,     // Array of tokens (without comments and whitespace)
    allowVoid,  // Is void keyword allowed?
    offset      // Optional: Index tokens of token in question
) {
    const index = offset ? offset : 0;
    if (index < tokens.length && tokens[index].type == JavaToken.TYPE_IDENTIFIER)
        return [tokens[index].value, index + 1];
    const allowed = ['boolean', 'double', 'byte', 'int', 'short', 'char', 'long', 'float'];
    if (allowVoid)
        allowed.push('void');
    if (index < tokens.length && tokens[index].type == JavaToken.TYPE_KEYWORD && allowed.indexOf(tokens[index].value) != -1)
        return [tokens[index].value, index + 1];
    return [undefined, index];
}

/**
* Read modifiers (like visibility specifier, abstract or strictfp) from tokens (if available)
* @return Array     ["Object specifiing read values and tokens", "Index of next token to read"]
*/
JavaPart.modifiersFromTokens = function(
    tokens,     // Array of tokens (without comments and whitespace)
    allow,      // Sub array of ["visibility", "abstract", "synchronized", "transient", "final", "static", "strictfp", "volatile", "native"]
    offset      // Optional: Index tokens of token in question
) {
    const boolModifiers = ["abstract", "synchronized", "transient", "final", "static", "strictfp", "volatile", "native"];
    const boolModifiersAllowed = allow.filter(key => (boolModifiers.indexOf(key) != -1));
    const boolModifiersDisallowed = boolModifiers.filter(key => (boolModifiersAllowed.indexOf(key) == -1));
    const visibilityAllowed = (allow.indexOf('visibility') != -1);
    const result = {
        visibility: undefined,
        tokenVisibility: null,
        illegal: [],
        double: []
    };
    if (visibilityAllowed) {
        result['visibility'] = JavaPart.VISIBILITY_PACKAGE;
        result['tokenVisibility'] = null;
    }
    const nameToken = (key => 'token' + key.substring(0, 1).toUpperCase() + key.substring(1));
    boolModifiersAllowed.forEach(key => {
        result[key] = false;
        result[nameToken(key)] = null;
    });
    for (let index = offset ? offset : 0; true; ++index) {
        const token = tokens[index];
        if (index >= tokens.length || token.type != JavaToken.TYPE_KEYWORD)
            return [result, index];
        if (visibilityAllowed && ['private', 'protected', 'public'].indexOf(token.value) != -1) {
            if (result.tokenVisibility)
                result.double.push(token);
            else {
                result.tokenVisibility = token;
                [result.visibility] = JavaPart.visibilityFromTokens(tokens, index);
            }
            continue;
        }
        if (boolModifiersAllowed.indexOf(token.value) != -1) {
            if (result[token.value])
                result.double.push(token);
            else {
                result[token.value] = true;
                result[nameToken(token.value)] = token;
            }
            continue;
        }
        if (boolModifiersDisallowed.indexOf(token.value) != -1) {
            result.illegal.push(token);
            continue;
        }
        return [result, index];
    }
}

/***[ JavaPartVarDeclaration.js ] *************************/
/** Class for describing content of a variable declaration in Java-like source code */
function JavaPartVarDeclaration(
    tokenStart,     // First token belonging to part
    localName,      // Name of variable
    type,           // Type of variable
    dimensions      // Optional: Number of dimensions
) {
    JavaPart.call(this, tokenStart);
    this.localName = localName;
    this.type = type;
    if (dimensions)
        this.dimensions = dimensions;
}
JavaPartVarDeclaration.prototype = Object.create(JavaPart.prototype);
JavaPartVarDeclaration.prototype.constructor = JavaPartVarDeclaration;

/** Name of variable */
JavaPartVarDeclaration.prototype.localName = undefined;

/** Type of variable */
JavaPartVarDeclaration.prototype.type = undefined;

/** Number of dimensions */
JavaPartVarDeclaration.prototype.dimensions = 0;

/** Is it a final attribute? */
JavaPartVarDeclaration.prototype.isFinal = false;

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartVarDeclaration.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    if (targetValues.part != 'variable')
        return { conformity: 0, goodFit: false, problems: ['Variablendeklaration darf es nicht geben'], instance: this };
    let checked = 0;
    let matches = 0;
    let goodFit = undefined;
    const problems = [];
    if (targetValues.name !== undefined) {
        ++checked;
        if (this.compareSingleValue('localName', targetValues.name)) {
            goodFit = true;
            ++matches;
        } else {
            goodFit = false;
            problems.push('Bei Variablendeklaration vergebener Name ' + this.localName + ' ist falsch');
        }
    }
    if (targetValues.type !== undefined) {
        ++checked;
        if (this.compareSingleValue('type', targetValues.type))
            ++matches;
        else
            problems.push('Datentyp bei Variablendeklaration ist falsch');
    }
    if (targetValues.dimensions !== undefined) {
        ++checked;
        if (this.compareSingleValue('dimensions', targetValues.dimensions))
            ++matches;
        else
            problems.push('Dimensionsanzahl bei Variablendeklaration ist falsch');
    }
    if (targetValues.final !== undefined) {
        ++checked;
        if (this.compareSingleValue('isFinal', targetValues.final))
            ++matches;
        else
            problems.push('Variablendeklaration' (targetValues.final ? ' muss' : ' darf nicht') + ' final sein');
    }
    return { conformity: matches / checked, goodFit, problems, instance: this };
}

/**
* Convert instance to human readable description
* @return String        Human readable description of instance
*/
JavaPartVarDeclaration.prototype.toString = function() {
    return 'Variablendeklaration';
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPartVarDeclaration.prototype.toJSON = function() {
    return {
        part: "variable",
        name: this.localName,
        type: this.type,
        dimensions: this.dimensions,
        final: this.isFinal
    };
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartVarDeclaration.fromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings,       // Array of warnings
    index,          // Optional: Index of first token to use (pervious tokens will be discarded if this pattern applies)
    allowVoid       // Optional: Allow type void?
) {
    let type, varname;
    let dimensions = 0;
    if (!index)
        index = 0;
    const tokenStart = tokens[index];
    const isFinal = (index < tokens.length && tokens[index].type == JavaToken.TYPE_KEYWORD && tokens[index].value == 'final');
    if (isFinal)
        ++index;
    [type, index] = JavaPart.typeFromTokens(tokens, allowVoid ? true : false, index);
    if (type === undefined)
        return null;
    while (index < tokens.length && tokens[index].type == JavaToken.TYPE_SEPARATOR && tokens[index].value == '[') {
        ++dimensions;
        ++index;
        if (index >= tokens.length || tokens[index].type != JavaToken.TYPE_SEPARATOR || tokens[index].value != ']')
            return null;
        ++index;
    }
    if (index >= tokens.length || tokens[index].type != JavaToken.TYPE_IDENTIFIER)
        return null;
    varname = tokens[index].value;
    ++index;
    while (index < tokens.length && tokens[index].type == JavaToken.TYPE_SEPARATOR && tokens[index].value == '[') {
        ++dimensions;
        ++index;
        if (index >= tokens.length || tokens[index].type != JavaToken.TYPE_SEPARATOR || tokens[index].value != ']')
            return null;
        ++index;
    }
    const obj = new JavaPartVarDeclaration(tokenStart, varname, type, dimensions);
    if (isFinal)
        obj.isFinal = true;
    obj.tokenEnd = tokens[index - 1];
    tokens.splice(0, index);
    return obj;
}

/***[ JavaPartAttributeDeclaration.js ] *************************/
/** Class for describing content of an attribute declaration in Java-like source code */
function JavaPartAttributeDeclaration(
    tokenStart,     // First token belonging to part
    localName,      // Name of variable
    type,           // Type of variable
    visibility,     // Visibility of attribute (see JavaPart.VISIBILITY_*)
    dimensions      // Optional: Number of dimensions
) {
    JavaPart.call(this, tokenStart);
    this.localName = localName;
    this.type = type;
    if (dimensions)
        this.dimensions = dimensions;
}
JavaPartAttributeDeclaration.prototype = Object.create(JavaPart.prototype);
JavaPartAttributeDeclaration.prototype.constructor = JavaPartAttributeDeclaration;

/** Name of attribute */
JavaPartAttributeDeclaration.prototype.localName = undefined;

/** Type of attribute */
JavaPartAttributeDeclaration.prototype.type = undefined;

/** Number of dimensions */
JavaPartAttributeDeclaration.prototype.dimensions = 0;

/** Visibility of class (see JavaPart.VISIBILITY_*) */
JavaPartAttributeDeclaration.prototype.visibility = JavaPart.VISIBILITY_PACKAGE;

/** Is attribute static? */
JavaPartAttributeDeclaration.prototype.isStatic = false;

/** Is it a final attribute? */
JavaPartAttributeDeclaration.prototype.isFinal = false;

/** Containing tokens forming intitial expression */
JavaPartAttributeDeclaration.prototype.content = null;

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartAttributeDeclaration.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    if (targetValues.part != 'attribute')
        return { conformity: 0, goodFit: false, problems: ['Attribut ' + this.localName + ' darf es nicht geben'], instance: this };
    let checked = 0;
    let matches = 0;
    let goodFit = undefined;
    let instance = this;
    const problems = [];
    if (targetValues.name !== undefined) {
        ++checked;
        if (this.compareSingleValue('localName', targetValues.name)) {
            goodFit = true;
            ++matches;
        } else {
            goodFit = false;
            problems.push('Name des Attributs ' + this.localName + ' ist falsch');
        }
    }
    if (targetValues.type !== undefined) {
        ++checked;
        if (this.compareSingleValue('type', targetValues.type))
            ++matches;
        else
            problems.push('Datentyp von Attribut ' + this.localName + ' ist falsch');
    }
    if (targetValues.dimensions !== undefined) {
        ++checked;
        if (this.compareSingleValue('dimensions', targetValues.dimensions))
            ++matches;
        else
            problems.push('Dimensionsanzahl von Attribut ' + this.localName + ' ist falsch');
    }
    if (targetValues.visibility !== undefined) {
        ++checked;
        if (this.compareSingleValue('visibility', targetValues.visibility))
            ++matches;
        else
            problems.push('Sichtbarkeit von Attribut ' + this.localName + ' ist falsch');
    }
    if (targetValues.static !== undefined) {
        ++checked;
        if (this.compareSingleValue('isStatic', targetValues.static))
            ++matches;
        else
            problems.push('Attribut ' + this.localName + (targetValues.static ? ' muss' : ' darf nicht') + ' statisch sein');
    }
    if (targetValues.final !== undefined) {
        ++checked;
        if (this.compareSingleValue('isFinal', targetValues.final))
            ++matches;
        else
            problems.push('Attribut ' + this.localName + (targetValues.final ? ' muss' : ' darf nicht') + ' final sein');
    }
    if (targetValues.content !== undefined) {
        ++checked;
        const contentCheckResult = this.comparePlainContent(targetValues.content);
        matches += contentCheckResult.conformity;
        if (contentCheckResult.conformity < 1 && contentCheckResult.instance);
            instance = contentCheckResult.instance;
        while (contentCheckResult.problems.length)
            problems.push('Bei Attribut ' + this.localName + ': ' + contentCheckResult.problems.shift());
    }
    return { conformity: matches / checked, goodFit, problems, instance };
}

/**
* Convert instance to human readable description
* @return String        Human readable description of instance
*/
JavaPartAttributeDeclaration.prototype.toString = function() {
    return 'Attribut ' + this.localName;
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPartAttributeDeclaration.prototype.toJSON = function() {
    return {
        part: "attribute",
        name: this.localName,
        type: this.type,
        dimensions: this.dimensions,
        visibility: this.visibility,
        static: this.isStatic,
        final: this.isFinal,
        content: (this.content instanceof Array) ? this.content.reduce((prev, token) => prev + ' ' + token.toString(), '').trim() : false
    };
}

/**
* Create JavaPartAttributeDeclaration instance using information of JavaPartVarDeclaration instance `varInfos`
* @return JavaPartAttributeDeclaration      Created instance
*/
JavaPartAttributeDeclaration.fromVar = function(
    varInfos        // JavaPartVarDeclaration instance being template of instance to create
) {
    const isValid = (varInfos instanceof JavaPartVarDeclaration);
    const obj = new JavaPartAttributeDeclaration(
        isValid ? varInfos.tokenStart : null,
        isValid ? varInfos.localName : undefined,
        isValid ? varInfos.type : 'void',
        undefined,
        isValid ? varInfos.dimensions : undefined
    );
    if (isValid && (varInfos instanceof JavaPart))
        obj.tokenEnd = varInfos.tokenEnd;
    return obj;
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartAttributeDeclaration.fromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    const tokenStart = tokens[0];
    let [modifiers, index] = JavaPart.modifiersFromTokens(tokens, ['visibility', 'static', 'final', 'transient', 'volatile']);
    let type, varname;
    const varInfos = JavaPartVarDeclaration.fromTokens(tokens, warnings, index);
    if (!varInfos)
        return null;
    const obj = JavaPartAttributeDeclaration.fromVar(varInfos);
    obj.tokenStart = tokenStart;
    modifiers.illegal.forEach(token => warnings.push({ desc: 'Schlüsselwort ' + token.value + ' ist nicht zu Beginn einer Attributdeklaration erlaubt', token }));
    modifiers.double.forEach(token => warnings.push({ desc: 'Ein Schlüsselwort wie ' + token.value + ' darf nicht doppelt in Attributdeklaration stehen', token }));
    if (modifiers.visibility != obj.visibility)
        obj.visibility = modifiers.visibility;
    if (modifiers.static)
        obj.isStatic = true;
    if (modifiers.final)
        obj.isFinal = true;
    if (tokens.length && tokens[0].type == JavaToken.TYPE_OPERATOR && tokens[0].value == '=') {
        obj.content = [];
        while (tokens.length && !(tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == ';'))
            obj.content.push(tokens.shift());
        obj.tokenEnd = obj.content[obj.content.length - 1];
        obj.content.shift();
    }
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_SEPARATOR || tokens[0].value != ';')
        warnings.push({ desc: 'Semikolon „;“ fehlt nach Attributdeklaration', token: tokens[0] });
    return obj;
}

/***[ JavaPartMethodDeclaration.js ] *************************/
/** Class for describing a method declaration in Java-like source code */
function JavaPartMethodDeclaration(
    tokenStart,     // First token belonging to part
    localName,      // Name of variable
    type,           // Type of variable (or true if constructor)
    visibility,     // Visibility of attribute (see JavaPart.VISIBILITY_*)
    dimensions      // Optional: Number of dimensions
) {
    JavaPart.call(this, tokenStart);
    this.params = [];
    this.localName = localName;
    this.type = (type === true) ? false : type;
    if (dimensions)
        this.dimensions = dimensions;
}
JavaPartMethodDeclaration.prototype = Object.create(JavaPart.prototype);
JavaPartMethodDeclaration.prototype.constructor = JavaPartMethodDeclaration;

/** Name of method */
JavaPartMethodDeclaration.prototype.localName = undefined;

/** Return type of method (false if constructor) */
JavaPartMethodDeclaration.prototype.type = undefined;

/** Number of dimensions of return type */
JavaPartMethodDeclaration.prototype.dimensions = 0;

/** Visibility of class (see JavaPart.VISIBILITY_*) */
JavaPartMethodDeclaration.prototype.visibility = JavaPart.VISIBILITY_PACKAGE;

/** Parameter list (array of JavaPartVarDeclaration or JavaPartFinalVarDeclaration instances) */
JavaPartMethodDeclaration.prototype.params = [];

/** Is method static? */
JavaPartMethodDeclaration.prototype.isStatic = false;

/** Is it a final method? */
JavaPartMethodDeclaration.prototype.isFinal = false;

/** Is it an abstract method? */
JavaPartMethodDeclaration.prototype.isAbstract = false;

/** Inner sequence */
JavaPartMethodDeclaration.prototype.content = null;

/**
* Customize the source code to this object
* @return String    Customized code
*/
JavaPartMethodDeclaration.prototype.adaptTargetCode = function(
    code
) {
    code = '' + code;
    const specialchar = code.charAt(0).replace(/[.*+?^${}()|[\]\\]/, '\\$&');
    code = code.substring(1);
    if (specialchar == ' ')
        return code;
    code = code.replace(new RegExp(specialchar + '0', 'g'), this.localName);
    const fktReplace = (code, param, index) => code.replace(new RegExp(specialchar + (index + 1), 'g'), param.localName);
    code = this.params.slice(10).reduce(fktReplace, code);
    code = this.params.slice(0, 10).reduce(fktReplace, code);
    return code;
}

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartMethodDeclaration.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    if (['method', 'constructor'].indexOf(targetValues.part) == -1)
        return { conformity: 0, goodFit: false, problems: [this + ' darf es nicht geben'], instance: this };
    const targetIsConstructor = ('constructor' == targetValues.part);
    let checked = 0;
    let matches = 0;
    let goodFit = undefined;
    const problems = [];
    ++checked;
    if (targetIsConstructor != (this.type === false)) {
        problems.push(
            targetIsConstructor
            ? 'Die Methode ' + this.localName + ' müsste ein Konstruktor sein'
            : 'Der Konstruktor müsste eine normale Methode sein'
        );
    } else {
        ++matches;
        if (targetIsConstructor)
            goodFit = true;
    }
    if (targetValues.name !== undefined) {
        ++checked;
        if (this.compareSingleValue('localName', targetValues.name)) {
            if (!targetIsConstructor)
                goodFit = true;
            ++matches;
        } else {
            goodFit = false;
            problems.push('Name der Methode ' + this.localName + ' ist falsch');
        }
    }
    if (targetValues.dimensions !== undefined) {
        ++checked;
        if (this.compareSingleValue('dimensions', targetValues.dimensions)) {
            ++matches;
        } else {
            goodFit = false;
            if (targetValues.dimensions)
                problems.push('Rückgabewert der Methode ' + this.localName + ' muss ein ' + targetValues.dimensions + 'D-Array sein');
            else
                problems.push('Rückgabewert der Methode ' + this.localName + ' darf kein Array sein');
        }
    }
    [
        { internalName: 'isAbstract', targetName: 'abstract', desc: 'abstrakt', bool: true },
        { internalName: 'isStatic', targetName: 'static', desc: 'statisch', bool: true },
        { internalName: 'isFinal', targetName: 'final', desc: 'final', bool: true },
        { internalName: 'visibility', targetName: 'visibility', desc: 'Sichtbarkeit' },
        { internalName: 'type', targetName: 'type', desc: 'Datentyp des Rückgabewerts' }
    ].forEach(({ internalName, targetName, desc, bool }) => {
        if (targetValues[targetName] === undefined)
            return;
        ++checked;
        if (this.compareSingleValue(internalName, targetValues[targetName]))
            ++matches;
        else if (bool)
            problems.push(this + (targetValues[targetName] ? ' muss ' : ' darf nicht ') + desc + ' sein');
        else
            problems.push(desc + ' bei ' + this + ' ist falsch');
    });
    let instance = (matches < checked) ? this : null;
    if (targetValues.params instanceof Array) {
        const isAlternatives = targetValues.params.reduce((prev, curr) => (prev && (curr instanceof Array)), targetValues.params.length > 0);
        const alternatives = isAlternatives ? targetValues.params : [targetValues.params];
        const paramCheckResults = alternatives.map(alternative => {
            const result = this.compareWithTargetParams(alternative);
            result._weight = alternative.length + 1;
            return result;
        });
        paramCheckResults.sort((a, b) => (b.conformity - a.conformity));
        const checkResult = paramCheckResults.length ? paramCheckResults[0] : { conformity: 0, problems: ['Unbekannte Abweichung'], '_weight': 0 };
        checked += checkResult._weight;
        matches += checkResult._weight * checkResult.conformity;
        if (!instance && checkResult.problems.length)
            instance = checkResult.instance || this;
        checkResult.problems.forEach(problem => problems.push(problem));
    }
    if (targetValues.content !== undefined) {
        ++checked;
        const contentCheckResult = this.comparePlainContent(targetValues.content);
        matches += contentCheckResult.conformity;
        if (problems.length == 0 && contentCheckResult.problems.length > 0 && contentCheckResult.instance)
            instance = contentCheckResult.instance;
        while (contentCheckResult.problems.length)
            problems.push('Innere Sequenz von ' + this.localName + ': ' + contentCheckResult.problems.shift());
    }
    return { conformity: matches / checked, goodFit, problems, instance };
}

/**
* Compare params with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartMethodDeclaration.prototype.compareWithTargetParams = function(
    targetValues        // Object describing target params
) {
    let checked = 0;
    let matches = 0;
    const problems = [];
    let instance = undefined;
    for (let n = 0; n < Math.min(this.params.length, targetValues.length); ++n) {
        const checkResult = this.params[n].compareWithTargetValues(targetValues[n]);
        ++checked;
        matches += checkResult.conformity;
        checkResult.problems.forEach(problem => problems.push(problem.replace('Variablendeklaration', 'Parameter ' + (n + 1) + ' von ' + this)));
        if (checkResult.problems.length)
            instance = instance || checkResult.instance;
    }
    for (let n = this.params.length; n < targetValues.length; ++n) {
        ++checked;
        problems.push('Parameter ' + (n + 1) + ' von ' + this + ' fehlt');
    }
    for (let n = targetValues.length; n < this.params.length; ++n) {
        ++checked;
        problems.push('Parameter ' + (n + 1) + ' von ' + this + ' (' + this.params[n].localName + ') muss weg');
    }
    if (problems.length && !instance)
        instance = this;
    return { conformity: matches / Math.max(1, checked), goodFit: undefined, problems, instance };
}

/**
* Convert instance to human readable description
* @return String        Human readable description of instance
*/
JavaPartMethodDeclaration.prototype.toString = function() {
    return ((this.type === false) ? (this.isStatic ? 'Statische Initialisierung' : 'Konstruktor') : ('Methode ' + this.localName));
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPartMethodDeclaration.prototype.toJSON = function() {
    const data = {
        visibility: this.visibility,
        params: this.params.map(param => param.toJSON()),
        static: this.isStatic
    };
    if (this.type === false) {
        data.part = 'constructor';
    } else {
        data.part = 'method';
        data.name = this.localName;
        data.type = this.type;
        data.dimensions = this.dimensions;
        data.final = this.isFinal;
        data.abstract = this.isAbstract
    }
    if (!data.abstract)
        data.content = (this.content instanceof Array) ? this.content.reduce((prev, token) => prev + ' ' + token.toString(), '') : false;
    return data;
}

/**
* Read all content until closing `}` to `objMethod.content` respecting nested brackets
* @return boolean       Found closing bracket?
*/
JavaPartMethodDeclaration._skipInnerSequence = function(
    objMethod,      // JavaPartMethodDeclaration instance to asigne content
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    let token = objMethod.tokenEnd;
    let level = 1;
    objMethod.content = [];
    while (tokens.length) {
        token = tokens.shift();
        objMethod.content.push(token);
        if (token.type != JavaToken.TYPE_SEPARATOR)
            continue;
        switch (token.value) {
            case '{':
                ++level;
                continue;
            case '}':
                --level;
                if (!level) {
                    objMethod.tokenEnd = token;
                    objMethod.content.pop();
                    return true;
                }
                continue;
        }
    }
    warnings.push({ desc: 'Schließende geschweifte Klammer fehlt', token: null });
    obj.tokenEnd = token;
    return false;
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartMethodDeclaration.fromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings,       // Array of warnings
    className       // Name of class containing method
) {
    if (!tokens.length)
        return null;
    const tokenStart = tokens[0];
    // Just static initialisation?
    if (tokenStart.value == 'static' && tokenStart.type == JavaToken.TYPE_KEYWORD && tokens.length > 1 && tokens[1].value == '{' && tokens[1].type == JavaToken.TYPE_SEPARATOR) {
        const objStaticInit = new JavaPartMethodDeclaration(tokenStart, className, false, undefined, 0);
        objStaticInit.isStatic = true;
        tokens.shift();
        objStaticInit.tokenEnd = tokens.shift();
        JavaPartMethodDeclaration._skipInnerSequence(objStaticInit, tokens, warnings);
        return objStaticInit;
    }

    let [modifiers, index] = JavaPart.modifiersFromTokens(tokens, ['visibility', 'abstract', 'static', 'final', 'strictfp', 'synchronized', 'native']);
    let varInfos = null;
    if (index + 1 < tokens.length && tokens[index].type == JavaToken.TYPE_IDENTIFIER && tokens[index].value == className && tokens[index + 1].type == JavaToken.TYPE_SEPARATOR && tokens[index + 1].value == '(') {
        varInfos = { localName: className, type: true, varInfos: undefined };
        tokens.splice(0, index + 1);
    } else {
        let indexRoundBracket = index + 1;
        let inBrackets = 0;
        while (indexRoundBracket < tokens.length) {
            if (inBrackets && tokens[indexRoundBracket].type == JavaToken.TYPE_SEPARATOR && tokens[indexRoundBracket].value == ']') {
                ++indexRoundBracket;
                --inBrackets;
                continue;
            }
            if (tokens[indexRoundBracket].type == JavaToken.TYPE_SEPARATOR && tokens[indexRoundBracket].value == '[') {
                ++indexRoundBracket;
                ++inBrackets;
                continue;
            }
            if (!inBrackets)
                break;
            ++indexRoundBracket;
        }
        ++indexRoundBracket;
        if (indexRoundBracket >= tokens.length || tokens[indexRoundBracket].type != JavaToken.TYPE_SEPARATOR || tokens[indexRoundBracket].value != '(')
            return null;
    }
    if (!varInfos)
        varInfos = JavaPartVarDeclaration.fromTokens(tokens, warnings, index, true);
    if (!varInfos)
        return null;
    const obj = new JavaPartMethodDeclaration(tokenStart, varInfos.localName, varInfos.type, undefined, varInfos.dimensions);
    if (modifiers.visibility != obj.visibility)
        obj.visibility = modifiers.visibility;
    if (modifiers.abstract)
        obj.isAbstract = true;
    if (modifiers.static)
        obj.isStatic = true;
    if (modifiers.final)
        obj.isFinal = true;
    /*!*/
    tokens.shift();
    while (tokens.length && !(tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == ')')) {
        if (obj.params.length) {
            if (tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == ',')
                tokens.shift();
            else
                warnings.push({ desc: 'Komma erwartet', token: tokens[0] });
        }
        const param = JavaPartVarDeclaration.fromTokens(tokens, warnings);
        if (!param) {
            warnings.push({ desc: 'Parameterdeklaration oder schließende Klammer erwartet', token: tokens.shift() });
            continue;
        }
        obj.params.push(param);
    }
    if (!tokens.length) {
        warnings.push({ desc: 'Parameterdeklaration oder schließende Klammer erwartet', token: null });
        obj.tokenEnd = null;
        return obj;
    }
    tokens.shift();
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_SEPARATOR || [';', '{'].indexOf(tokens[0].value) == -1) {
        warnings.push({ desc: modifiers.abstract ? 'Semikolon erwartet' : 'Innere Sequenz erwartet', token: tokens[0] });
        obj.tokenEnd = tokens[0];
        return obj;
    }
    if (modifiers.abstract && tokens[0].value == ';') {
        obj.tokenEnd = tokens.shift();
        return obj;
    }
    if (!modifiers.abstract && tokens[0].value == ';') {
        warnings.push({ desc: 'Statt Semikolon bei nicht-abstrakter Methode innere Sequenz erwartet', token: tokens[0] });
        obj.tokenEnd = tokens.shift();
        return obj;
    }
    if (modifiers.abstract && tokens[0].value == '{')
        warnings.push({ desc: 'Bei abstrakter Methode kann keine Sequenz vorgegeben werden', token: tokens[0] });
    // Skip inner sequence
    tokens.shift();
    JavaPartMethodDeclaration._skipInnerSequence(obj, tokens, warnings);
    return obj;
}

/***[ JavaPartClassDeclaration.js ] *************************/
/** Class for describing content of a class declaration in Java-like source code */
function JavaPartClassDeclaration(
    tokenStart      // First token belonging to part
) {
    JavaPart.call(this, tokenStart);
    this.implements = [];
    this.content = [];
}
JavaPartClassDeclaration.prototype = Object.create(JavaPart.prototype);
JavaPartClassDeclaration.prototype.constructor = JavaPartClassDeclaration;

/*!*/// Not implemented: final, strictfp

/** Name of declared class */
JavaPartClassDeclaration.prototype.className = undefined;

/** Name of parent class (if any) */
JavaPartClassDeclaration.prototype.parentClass = 'Object';

/** Name of interfaces implemented by class */
JavaPartClassDeclaration.prototype.implements = [];

/** Is class abstract? */
JavaPartClassDeclaration.prototype.isAbstract = false;

/** Is class static? */
JavaPartClassDeclaration.prototype.isStatic = false;

/** Is it a final class? */
JavaPartClassDeclaration.prototype.isFinal = false;

/** Visibility of class (see JavaPart.VISIBILITY_*) */
JavaPartClassDeclaration.prototype.visibility = JavaPart.VISIBILITY_PACKAGE;

/** Contained definitions */
JavaPartClassDeclaration.prototype.content = [];

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartClassDeclaration.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    if (targetValues.part != 'class')
        return { conformity: 0, goodFit: false, problems: ['Klasse ' + this.className + ' darf es nicht geben'], instance: this };
    let checked = 0;
    let matches = 0;
    let goodFit = undefined;
    const problems = [];
    if (targetValues.name !== undefined) {
        ++checked;
        if (this.compareSingleValue('className', targetValues.name)) {
            goodFit = true;
            ++matches;
        } else {
            goodFit = false;
            problems.push('Name der Klasse ' + this.className + ' ist falsch');
        }
    }
    [
        { internalName: 'isAbstract', targetName: 'abstract', desc: 'abstrakt', bool: true },
        { internalName: 'isStatic', targetName: 'static', desc: 'statisch', bool: true },
        { internalName: 'isFinal', targetName: 'final', desc: 'final', bool: true },
        { internalName: 'parentClass', targetName: 'extends', desc: 'Elternklasse' },
        { internalName: 'visibility', targetName: 'visibility', desc: 'Sichtbarkeit' },
        { internalName: 'implements', targetName: 'implements', desc: 'Auflistung implementierter Schnittstellen' }
    ].forEach(({ internalName, targetName, desc, bool }) => {
        if (targetValues[targetName] === undefined)
            return;
        ++checked;
        if (this.compareSingleValue(internalName, targetValues[targetName]))
            ++matches;
        else if (bool)
            problems.push('Klasse ' + this.className + (targetValues[targetName] ? 'muss ' : 'darf nicht ') + desc + ' sein');
        else
            problems.push(desc + ' bei Klasse ' + this.className + ' ist falsch');
    });
    let instance = (matches < checked) ? this : null;
    if (targetValues.content instanceof Array) {
        const contentCheckResult = JavaPart.compareContent(this.content, targetValues.content);
        const weight = targetValues.content.length + 1;
        checked += weight;
        matches += weight * contentCheckResult.conformity;
        if (!instance && contentCheckResult.problems.length)
            instance = contentCheckResult.instance || this;
        contentCheckResult.problems.forEach(problem => problems.push('In ' + this + ': ' + problem));
    }
    return { conformity: matches / checked, goodFit, problems, instance };
}

/**
* Convert instance to human readable description
* @return String        Human readable description of instance
*/
JavaPartClassDeclaration.prototype.toString = function() {
    return 'Klasse ' + this.className;
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPartClassDeclaration.prototype.toJSON = function() {
    return {
        part: 'class',
        name: this.className,
        extends: this.parentClass,
        implements: this.implements.slice(),
        abstract: this.isAbstract,
        final: this.isFinal,
        static: this.isStatic,
        content: this.content.map(part => part.toJSON())
    };
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartClassDeclaration._tryToStartFromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    const tokenStart = tokens[0];
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_KEYWORD)
        return null;
    let [modifiers, index] = JavaPart.modifiersFromTokens(tokens, ['visibility', 'abstract', 'static', 'final', 'strictfp']);
    if (index >= tokens.length || tokens[index].type != JavaToken.TYPE_KEYWORD || tokens[index].value != 'class')
        return null;
    const instance = new JavaPartClassDeclaration(tokenStart);
    modifiers.illegal.forEach(token => warnings.push({ desc: 'Schlüsselwort ' + token.value + ' ist nicht zu Beginn einer Klassendefinition erlaubt', token }));
    modifiers.double.forEach(token => warnings.push({ desc: 'Ein Schlüsselwort wie ' + token.value + ' darf nicht doppelt in Klassendefinition stehen', token }));
    if (modifiers.abstract)
        instance.isAbstract = true;
    if (modifiers.static)
        instance.isStatic = true;
    if (modifiers.final)
        instance.isFinal = true;
    if (modifiers.visibility !== instance.visibility) {
        if (modifiers.visibility == JavaPart.VISIBILITY_PROTECTED)
            warnings.push({ desc: 'Die Sichtbarkeit einer Klasse kann nicht protected sein', token: modifiers.tokenVisibility });
        else if (modifiers.visibility == JavaPart.VISIBILITY_PRIVATE)
            warnings.push({ desc: 'Die Sichtbarkeit einer Klasse kann nicht private sein', token: modifiers.tokenVisibility });
        else
            instance.visibility = modifiers.visibility;
    }
    tokens.splice(0, index + 1);
    return instance;
}

/**
* Read class name from tokens and apply name to `obj`
* @return undefined
*/
JavaPartClassDeclaration._readNameFromTokens = function(
    obj,            // JavaPartClassDeclaration instance beeing read from tokens
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_IDENTIFIER) {
        warnings.push({ desc: 'Nach `class` muss der Klassenname stehen', token: tokens[0] });
        return;
    }
    obj.className = tokens.shift().value;
}

/**
* Read parent class name from tokens (if available)
* @return boolean       Is parent class specified?
*/
JavaPartClassDeclaration._parentClassFromTokens = function(
    obj,            // JavaPartClassDeclaration instance beeing read from tokens
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_KEYWORD || tokens[0].value != 'extends')
        return false;
    tokens.shift();
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_IDENTIFIER) {
        warnings.push({ desc: 'Nach `extends` muss ein Klassenname stehen', token: tokens[0] });
        return true;
    }
    obj.parentClass = tokens.shift().value;
    return true;
}

/**
* Read implemented interfaces from tokens (if available)
* @return boolean       Is there no list of implemented interfaces or a valid one?
*/
JavaPartClassDeclaration._interfacesFromTokens = function(
    obj,            // JavaPartClassDeclaration instance beeing read from tokens
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_KEYWORD || tokens[0].value != 'implements')
        return true;
    tokens.shift();
    let needComma = false;
    while (tokens.length) {
        if (needComma && tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == '{')
            return true;
        if (needComma && tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == ',') {
            needComma = false;
            tokens.shift();
            continue;
        }
        if (!needComma && tokens[0].type == JavaToken.TYPE_IDENTIFIER) {
            needComma = true;
            if (obj.implements.indexOf(tokens[0].value) == -1)
                obj.implements.push(tokens[0].value);
            else
                warnings.push({ desc: 'Derselbe Interface-Name darf nicht doppelt aufgeführt werden', token: tokens[0] });
            tokens.shift();
            continue;
        }
        break;
    }
    if (needComma)
        warnings.push({ desc: 'Öffnende geschweifte Klammer oder Komma fehlt', token: tokens[0] });
    else
        warnings.push({ desc: 'Interface-Name fehlt', token: tokens[0] });
    return false;
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartClassDeclaration.fromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    const obj = JavaPartClassDeclaration._tryToStartFromTokens(tokens, warnings);
    if (!obj)
        return obj;
    JavaPartClassDeclaration._readNameFromTokens(obj, tokens, warnings);
    JavaPartClassDeclaration._parentClassFromTokens(obj, tokens, warnings);
    if (!JavaPartClassDeclaration._interfacesFromTokens(obj, tokens, warnings))
        return obj;
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_SEPARATOR || tokens[0].value != '{') {
        warnings.push({ desc: 'Öffnende geschweifte Klammer fehlt', token: tokens[0] });
        return obj;
    }
    tokens.shift();

    function shiftSemicolon() {
        if (!tokens.length || tokens[0].type != JavaToken.TYPE_SEPARATOR || tokens[0].value != ';')
            return false;
        tokens.shift();
        return true;
    }

    let content;
    while(true) {
        while (shiftSemicolon())
            ;
        content = (JavaPartMethodDeclaration.fromTokens(tokens, warnings, obj.className) || JavaPartAttributeDeclaration.fromTokens(tokens, warnings));
        if (!content)
            break;
        obj.content.push(content);
    }

    while (shiftSemicolon())
        ;
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_SEPARATOR || tokens[0].value != '}')
        warnings.push({ desc: 'Schließende geschweifte Klammer fehlt' + (tokens.length ? ' oder das Nachfolgende ist fehlerhaft aufgebaut' : ''), token: tokens[0] });
    obj.tokenEnd = tokens.shift();
    return obj;
}

/***[ JavaPartImport.js ] *************************/
/** Class for describing content of an import statement in Java-like source code */
function JavaPartImport(
    tokenStart      // First token belonging to part
) {
    JavaPart.call(this, tokenStart);
    this.implements = [];
    this.content = [];
}
JavaPartImport.prototype = Object.create(JavaPart.prototype);
JavaPartImport.prototype.constructor = JavaPartImport;

/** Path of ressource to import */
JavaPartImport.prototype.path = [];
Object.defineProperty(JavaPartImport.prototype, 'pathString', {
    enumerable: true,
    get() { return this.path.join('.'); }
});

/** Is import static? */
JavaPartImport.prototype.isStatic = false;

/**
* Convert instance to human readable description
* @return String        Human readable description of instance
*/
JavaPartImport.prototype.toString = function() {
    if (!this.path.length)
        return 'Leere Importanweisung';
    return 'Import von ' + this.pathString;
}

/**
* Compare this instance with some target values
* @return Object    { conformity: "Proportion of values equaling target values", goodFit: "Does for instance the name fit?", problems: "Array of mismatches", instance: "Mismatching (sub-)instance" }
*/
JavaPartImport.prototype.compareWithTargetValues = function(
    targetValues        // Object describing target values
) {
    if (targetValues.part != 'import')
        return { conformity: 0, goodFit: false, problems: [this.toString() + ' darf es nicht geben'], instance: this };
    let checked = 0;
    let matches = 0;
    let goodFit = undefined;
    const problems = [];
    if (targetValues.path !== undefined) {
        ++checked;
        if (this.compareSingleValue('pathString', targetValues.path)) {
            goodFit = true;
            ++matches;
        } else {
            goodFit = false;
            problems.push(this.path.join('.') + ' ist nicht das, was importiert werden muss');
        }
    }

    if (targetValues.static !== undefined) {
        ++checked;
        if (this.compareSingleValue('isStatic', targetValues.static))
            ++matches;
        else
            problems.push(this.toString() + (targetValues[targetName] ? 'muss ' : 'darf nicht ') + 'statisch sein');
    }
    return { conformity: matches / checked, goodFit, problems, instance };
}

/**
* Convert data of this object to JSON data
* @return Object        JSON data
*/
JavaPartImport.prototype.toJSON = function() {
    return {
        part: 'import',
        path: this.pathString,
        static: this.isStatic
    };
}

/**
* Convert first tokens to JavaPart instance if possible
* @return JavaPart      Created JavaPart instance (or null)
*/
JavaPartImport.fromTokens = function(
    tokens,         // Array of tokens (without comments and whitespace)
    warnings        // Array of warnings
) {
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_KEYWORD || tokens[0].value != 'import')
        return null;
    const obj = new JavaPartImport(tokens.shift());
    let pathEnded = false;
    if (tokens.length && tokens[0].type == JavaToken.TYPE_KEYWORD && tokens[0].value == 'static') {
        obj.isStatic = true;
        obj.tokenEnd = tokens.shift();
    }
    if (!tokens.length || tokens[0].type != JavaToken.TYPE_IDENTIFIER) {
        warnings.push({ desc: 'Angabe fehlt, was importiert werden soll', token: tokens[0] });
        return obj;
    }
    obj.path.push(tokens.shift().value);
    while (true) {
        if (!tokens.length) {
            warnings.push({ desc: 'Importanweisung muss mit Semikolon enden', token: tokens[0] });
            return obj;
        }
        if (tokens[0].type == JavaToken.TYPE_SEPARATOR && tokens[0].value == ';') {
            obj.tokenEnd = tokens.shift();
            return obj;
        }
        if (pathEnded) {
            warnings.push({ desc: 'Nach dem Sternchen muss bei der Importanweisung direkt ein Semikolon folgen', token: tokens[0] });
            obj.tokenEnd = tokens.shift();
            continue;
        }
        if (tokens[0].type != JavaToken.TYPE_SEPARATOR || tokens[0].value != '.') {
            warnings.push({ desc: 'Angabe in Importanweisung ergibt keinen Sinn', token: tokens[0] });
            return obj;
        }
        obj.tokenEnd = tokens.shift();
        if (!tokens.length) {
            warnings.push({ desc: 'Nach dem letzten Punkt in der Importanweisung fehlt ein Bezeichner', token: obj.tokenEnd });
            return obj;
        }
        if (tokens[0].type == JavaToken.TYPE_OPERATOR && tokens[0].value == '*') {
            if (!obj.path.length)
                warnings.push({ desc: 'Ein Sternchen zu Beginn einer Importanweisung ergibt keinen Sinn', token: tokens[0] });
            obj.path.push('*');
            obj.tokenEnd = tokens.shift();
            pathEnded = true;
            continue;
        }
        if (tokens[0].type == JavaToken.TYPE_IDENTIFIER) {
            obj.path.push(tokens[0].value);
            obj.tokenEnd = tokens.shift();
            continue;
        }
        warnings.push({ desc: 'Nach dem Punkt muss bei der Importanweisung direkt ein Bezeichner folgen', token: tokens[0] });
        obj.tokenEnd = tokens.shift();
    }
}
